/*
 *
 *  Copyright (C) 2000-2001, OFFIS
 *
 *  This software and supporting documentation were developed by
 *
 *    Kuratorium OFFIS e.V.
 *    Healthcare Information and Communication Systems
 *    Escherweg 2
 *    D-26121 Oldenburg, Germany
 *
 *  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
 *  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
 *  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
 *  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
 *  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
 *
 *  Module:  dcmsr
 *
 *  Author:  Joerg Riesmeier
 *
 *  Purpose: Renders the contents of a DICOM structured reporting file in
 *           HTML format
 *
 *  Last Update:      $Author: meichel $
 *  Update Date:      $Date: 2001/06/01 15:50:57 $
 *  Source File:      $Source: /share/dicom/cvs-depot/dcmtk/dcmsr/apps/dsr2html.cc,v $
 *  CVS/RCS Revision: $Revision: 1.8 $
 *  Status:           $State: Exp $
 *
 *  CVS/RCS Log at end of file
 *
 */


#include "osconfig.h"    /* make sure OS specific configuration is included first */

#include "dsrdoc.h"

#include "dcdebug.h"
#include "cmdlnarg.h"
#include "ofconapp.h"
#include "dcuid.h"    /* for dcmtk version name */

#include <iostream.h>
#include <fstream.h>


#define OFFIS_CONSOLE_APPLICATION "dsr2html"

static char rcsid[] = "$dcmtk: " OFFIS_CONSOLE_APPLICATION " v"
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";


// ********************************************


static E_Condition renderFile(ostream &out,
                              const char *ifname,
                              const char *cssName,
                              const OFBool isDataset,
                              const E_TransferSyntax xfer,
                              const size_t readFlags,
                              const size_t renderFlags,
                              const OFBool debugMode)
{
    E_Condition result = EC_Normal;

    if ((ifname == NULL) || (strlen(ifname) == 0))
    {
        CERR << OFFIS_CONSOLE_APPLICATION << ": invalid filename: <empty string>" << endl;
        return EC_IllegalCall;
    }

    DcmFileStream myin(ifname, DCM_ReadMode);
    if (myin.GetError() != EC_Normal)
    {
        CERR << OFFIS_CONSOLE_APPLICATION << ": cannot open file: " << ifname << endl;
        return EC_IllegalCall;
    }

    DcmObject *dfile = NULL;
    if (isDataset)
        dfile = new DcmDataset();
    else
        dfile = new DcmFileFormat();

    if (dfile != NULL)
    {
        dfile->transferInit();
        dfile->read(myin, xfer, EGL_noChange);
        dfile->transferEnd();

        if (dfile->error() != EC_Normal)
        {
            CERR << OFFIS_CONSOLE_APPLICATION << ": error: " << dcmErrorConditionToString(dfile->error())
                 << ": reading file: "<< ifname << endl;
            result = EC_IllegalCall;
        }
    } else
        result = EC_MemoryExhausted;

    if (result == EC_Normal)
    {
        result = EC_IllegalCall;
        DcmDataset *dset = (isDataset) ? (DcmDataset*)dfile : ((DcmFileFormat *)dfile)->getDataset();
        if (dset != NULL)
        {
            DSRDocument *dsrdoc = new DSRDocument();
            if (dsrdoc != NULL)
            {
                if (debugMode)
                    dsrdoc->setLogStream(&ofConsole);
                result = dsrdoc->read(*dset, readFlags);
                if (result == EC_Normal)
                    result = dsrdoc->renderHTML(out, renderFlags, cssName);
                else
                    CERR << OFFIS_CONSOLE_APPLICATION << ": error: parsing SR file: "<< ifname << endl;
            }
            delete dsrdoc;
        }
    }
    delete dfile;

    return result;
}


#define SHORTCOL 3
#define LONGCOL 21


int main(int argc, char *argv[])
{
    int opt_debugMode = 0;
    size_t opt_readFlags = 0;
    size_t opt_renderFlags = DSRTypes::HF_renderDcmtkFootnote;
    const char *opt_cssName = NULL;
    OFBool isDataset = OFFalse;
    E_TransferSyntax xfer = EXS_Unknown;

    SetDebugLevel(( 0 ));

    OFConsoleApplication app(OFFIS_CONSOLE_APPLICATION, "Render DICOM SR file and data set to HTML", rcsid);
    OFCommandLine cmd;
    cmd.setOptionColumns(LONGCOL, SHORTCOL);
    cmd.setParamColumn(LONGCOL + SHORTCOL + 4);

    cmd.addParam("dsrfile-in",   "DICOM SR input filename to be rendered", OFCmdParam::PM_Mandatory);
    cmd.addParam("htmlfile-out", "HTML output filename (default: stdout)", OFCmdParam::PM_Optional);

    cmd.addGroup("general options:", LONGCOL, SHORTCOL + 2);
      cmd.addOption("--help",                  "-h",     "print this help text and exit");
      cmd.addOption("--debug",                 "-d",     "debug mode, print debug information");

    cmd.addGroup("input options:");
      cmd.addSubGroup("input file format:");
        cmd.addOption("--read-file",           "+f",     "read file format or data set (default)");
        cmd.addOption("--read-dataset",        "-f",     "read data set without file meta information");
      cmd.addSubGroup("input transfer syntax (only with --read-dataset):");
        cmd.addOption("--read-xfer-auto",      "-t=",    "use TS recognition (default)");
        cmd.addOption("--read-xfer-little",    "-te",    "read with explicit VR little endian TS");
        cmd.addOption("--read-xfer-big",       "-tb",    "read with explicit VR big endian TS");
        cmd.addOption("--read-xfer-implicit",  "-ti",    "read with implicit VR little endian TS");

    cmd.addGroup("parsing options:");
      cmd.addSubGroup("error handling:");
        cmd.addOption("--ignore-constraints",  "-Ec",    "ignore relationship content constraints");

    cmd.addGroup("output options:");
      cmd.addSubGroup("HTML compatibility:");
        cmd.addOption("--html-3.2",            "+H3",    "use only HTML version 3.2 compatible features");
        cmd.addOption("--html-4.0",            "+H4",    "allow all HTML version 4.0 features (default)");
        cmd.addOption("--add-document-type",   "+Hd",    "add reference to SGML document type definition");
      cmd.addSubGroup("cascading style sheet (CSS), only with HTML 4.0:");
        cmd.addOption("--css-reference",       "+Sr", 1, "URL : string",
                                                         "add reference to specified CSS to HTML page");
        cmd.addOption("--css-file",            "+Sf", 1, "filename : string",
                                                         "embed content of specified CSS into HTML page");
      cmd.addSubGroup("general rendering:");
        cmd.addOption("--expand-inline",       "+Ri",    "expand short content items inline (default)");
        cmd.addOption("--never-expand-inline", "-Ri",    "never expand content items inline");
        cmd.addOption("--render-full-data",    "+Rd",    "render full data of content items");
      cmd.addSubGroup("document rendering:");
        cmd.addOption("--document-type-title", "+Dt",    "use document type as document title (default)");
        cmd.addOption("--patient-info-title",  "+Dp",    "use patient information as document title");
        cmd.addOption("--no-document-header",  "-Dh",    "do not render general document information");
      cmd.addSubGroup("code rendering:");
        cmd.addOption("--render-inline-codes", "+Ci",    "render codes in continuous text blocks");
        cmd.addOption("--concept-name-codes",  "+Cn",    "render code of concept names");
        cmd.addOption("--numeric-unit-codes",  "+Cu",    "render code of numeric measurement units");
        cmd.addOption("--code-value-unit",     "+Cv",    "use code value as measurement unit (default)");
        cmd.addOption("--code-meaning-unit",   "+Cm",    "use code meaning as measurement unit");
        cmd.addOption("--render-all-codes",    "+Ca",    "render all codes (+Ci, +Cn, +Cu)");

    /* evaluate command line */
    prepareCmdLineArgs(argc, argv, OFFIS_CONSOLE_APPLICATION);
    if (app.parseCommandLine(cmd, argc, argv, OFCommandLine::ExpandWildcards))
    {
        /* general options */
        if (cmd.findOption("--debug"))
            opt_debugMode = 5;

        /* input options */
        cmd.beginOptionBlock();
        if (cmd.findOption("--read-file"))
            isDataset = OFFalse;
        if (cmd.findOption("--read-dataset"))
            isDataset = OFTrue;
        cmd.endOptionBlock();

        cmd.beginOptionBlock();
        if (cmd.findOption("--read-xfer-auto"))
        {
            app.checkDependence("--read-xfer-auto", "--read-dataset", isDataset);
            xfer = EXS_Unknown;
        }
        if (cmd.findOption("--read-xfer-little"))
        {
            app.checkDependence("--read-xfer-little", "--read-dataset", isDataset);
            xfer = EXS_LittleEndianExplicit;
        }
        if (cmd.findOption("--read-xfer-big"))
        {
            app.checkDependence("--read-xfer-big", "--read-dataset", isDataset);
            xfer = EXS_BigEndianExplicit;
        }
        if (cmd.findOption("--read-xfer-implicit"))
        {
            app.checkDependence("--read-xfer-implicit", "--read-dataset", isDataset);
            xfer = EXS_LittleEndianImplicit;
        }
        cmd.endOptionBlock();

        if (cmd.findOption("--ignore-constraints"))
            opt_readFlags |= DSRTypes::RF_ignoreRelationshipConstraints;                    

        /* HTML compatibility */
        cmd.beginOptionBlock();
        if (cmd.findOption("--html-3.2"))
            opt_renderFlags |= DSRTypes::HF_version32Compatibility;
        if (cmd.findOption("--html-4.0"))
        {
            /* default */
        }
        cmd.endOptionBlock();

        if (cmd.findOption("--add-document-type"))
            opt_renderFlags |= DSRTypes::HF_addDocumentTypeReference;

        /* cascading style sheet */
        cmd.beginOptionBlock();
        if (cmd.findOption("--css-reference"))
        {
          	app.checkDependence("--css-reference", "--html-4.0", !(opt_renderFlags & DSRTypes::HF_version32Compatibility));
            opt_renderFlags &= ~DSRTypes::HF_copyStyleSheetContent;
            app.checkValue(cmd.getValue(opt_cssName));           
        }
        if (cmd.findOption("--css-file"))
        {
          	app.checkDependence("--css-file", "--html-4.0", !(opt_renderFlags & DSRTypes::HF_version32Compatibility));
            opt_renderFlags |= DSRTypes::HF_copyStyleSheetContent;
            app.checkValue(cmd.getValue(opt_cssName));           
        }
        cmd.endOptionBlock();

        /* general rendering */
        cmd.beginOptionBlock();
        if (cmd.findOption("--expand-inline"))
        {
            /* default */
        }
        if (cmd.findOption("--never-expand-inline"))
            opt_renderFlags |= DSRTypes::HF_neverExpandChildrenInline;
        cmd.endOptionBlock();

        if (cmd.findOption("--render-full-data"))
            opt_renderFlags |= DSRTypes::HF_renderFullData;

        /* document rendering */
        cmd.beginOptionBlock();
        if (cmd.findOption("--document-type-title"))
        {
            /* default */
        }
        if (cmd.findOption("--patient-info-title"))
            opt_renderFlags |= DSRTypes::HF_renderPatientTitle;
        cmd.endOptionBlock();

        if (cmd.findOption("--no-document-header"))
            opt_renderFlags |= DSRTypes::HF_renderNoDocumentHeader;

        /* code rendering */
        if (cmd.findOption("--render-inline-codes"))
            opt_renderFlags |= DSRTypes::HF_renderInlineCodes;
        if (cmd.findOption("--concept-name-codes"))
            opt_renderFlags |= DSRTypes::HF_renderConceptNameCodes;
        if (cmd.findOption("--numeric-unit-codes"))
            opt_renderFlags |= DSRTypes::HF_renderNumericUnitCodes;
        if (cmd.findOption("--code-value-unit"))
            opt_renderFlags &= ~DSRTypes::HF_useCodeMeaningAsUnit;
        if (cmd.findOption("--code-meaning-unit"))
            opt_renderFlags |= DSRTypes::HF_useCodeMeaningAsUnit;
        if (cmd.findOption("--render-all-codes"))
            opt_renderFlags |= DSRTypes::HF_renderAllCodes;
    }

    SetDebugLevel((opt_debugMode));

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded())
    {
        CERR << "Warning: no data dictionary loaded, "
             << "check environment variable: "
             << DCM_DICT_ENVIRONMENT_VARIABLE << endl;
    }

    int result = 0;
    const char *ifname = NULL;
    cmd.getParam(1, ifname);
    if (cmd.getParamCount() == 2)
    {
        const char *ofname = NULL;
        cmd.getParam(2, ofname);
        ofstream stream(ofname);
        if (stream.good())
        {
            if (renderFile(stream, ifname, opt_cssName, isDataset, xfer, opt_readFlags, opt_renderFlags, opt_debugMode != 0) != EC_Normal)
                result = 2;
        } else
            result = 1;
    } else {
        if (renderFile(COUT, ifname, opt_cssName, isDataset, xfer, opt_readFlags, opt_renderFlags, opt_debugMode != 0) != EC_Normal)
            result = 3;
    }

    return result;
}


/*
 * CVS/RCS Log:
 * $Log: dsr2html.cc,v $
 * Revision 1.8  2001/06/01 15:50:57  meichel
 * Updated copyright header
 *
 * Revision 1.7  2001/04/03 08:22:54  joergr
 * Added new command line option: ignore relationship content constraints
 * specified for each SR document class.
 *
 * Revision 1.6  2000/12/08 16:06:19  joergr
 * Replaced empty code lines (";") by empty command blocks ("{}") to avoid
 * compiler warnings reported by MSVC6.
 *
 * Revision 1.5  2000/11/09 20:31:08  joergr
 * Added new command line options (document type and HTML version).
 *
 * Revision 1.4  2000/11/07 18:09:48  joergr
 * Added new command line option allowing to choose code value or meaning to be
 * rendered as the numeric measurement unit.
 *
 * Revision 1.3  2000/11/01 16:08:04  joergr
 * Added support for Cascading Style Sheet (CSS) used optionally for HTML
 * rendering. Optimized HTML rendering.
 *
 * Revision 1.2  2000/10/26 14:15:33  joergr
 * Added new flag specifying whether to add a "dcmtk" footnote to the rendered
 * HTML document or not.
 *
 * Revision 1.1  2000/10/13 07:46:21  joergr
 * Added new module 'dcmsr' providing access to DICOM structured reporting
 * documents (supplement 23).  Doc++ documentation not yet completed.
 *
 *
 */
